package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.config.ApplicationProperties;
import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;
import at.ac.uibk.gitsearch.es.model.ExerciseInfo;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.dataformat.yaml.YAMLGenerator;
import org.apache.commons.io.FileUtils;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.lib.Repository;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.Group;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import java.io.*;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

@Service
public class ExerciseService {

    private final Logger log = LoggerFactory.getLogger(ExerciseService.class);

    private final ApplicationProperties applicationProperties;

    private final FileService fileService;

    private final GitlabService gitlabService;

    public ExerciseService(ApplicationProperties applicationProperties, FileService fileService, GitlabService gitlabService) {
        this.applicationProperties = applicationProperties;
        this.fileService = fileService;
        this.gitlabService = gitlabService;
    }

    /**
     * Retrieve a zipped Artemis Exercise from a given URL
     * and store it in 'tmp-import' directory
     * @param exerciseUrl the url of the exposed exercise
     * @throws ArtemisImportError
     */
    public void importExerciseFromArtemis(String exerciseUrl) throws ArtemisImportError{
        String exerciseToken = exerciseUrl.split("/")[exerciseUrl.split("/").length - 1];
        Client client = ClientBuilder.newClient();

        WebTarget target = client.target(exerciseUrl);
        try {
            InputStream inputStream = target.request().accept(javax.ws.rs.core.MediaType.APPLICATION_OCTET_STREAM).get(InputStream.class);

            ZipInputStream zipInput = new ZipInputStream(inputStream);
            storeZipInTmp(zipInput, exerciseToken);

        } catch (Exception e) {
            throw new ArtemisImportError("ZipFile couldn't be retrieved from [" + URLDecoder.decode(exerciseUrl, StandardCharsets.UTF_8) + "]" + e.getMessage());
        }
    }

    /**
     * Import exercise to Gitlab.
     * Sequentially executes following processes:
     * - Apply user defined exercise metadata
     * - Create repository in Gitlab's group defined by the user
     * - Locally checking out the new created repository
     * - Copy, commit and push files to the new repository
     *
     * @param exerciseInfo of the exercise to import
     * @param exerciseToken of the stored exercise (in '/tmp-import' directory) to import
     */
    public void importExercise(ArtemisExerciseInfo exerciseInfo, String exerciseToken, Integer gitlabGroupId) throws GitLabApiException, GitAPIException, IOException, ArtemisImportError {
        applyExerciseInfoChanges(exerciseInfo, exerciseToken);

        gitlabService.setJGitDefaultAuth();
        Group group = gitlabService.getGroupById(gitlabGroupId);
        String repoUrl = gitlabService.createRepository(group, exerciseInfo.getTitle());
        Repository repo = gitlabService.checkoutRepo(repoUrl);
        gitlabService.commitAndPushToRepo(repo, new File("tmp-import", exerciseToken));

        repo.close();
        try {
            FileUtils.deleteDirectory(new File("tmp-import", exerciseToken));
            FileUtils.deleteDirectory(repo.getDirectory().getParentFile());
        } catch (IOException e) {
            log.error("Could not remove directory: {} \n", e.getMessage());
        }
    }

    /**
     * Retrieves metadata from imported Artemis exercise
     * @param exerciseToken of the imported exercise
     * @return Content of exercise's 'metadata.yaml' and 'artemis.yaml'
     */
    public ArtemisExerciseInfo getArtemisExerciseInfo(String exerciseToken) throws IOException {
        ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.enable(DeserializationFeature.READ_ENUMS_USING_TO_STRING)
            .enable(MapperFeature.ACCEPT_CASE_INSENSITIVE_ENUMS);
        ArtemisExerciseInfo artemisExerciseInfo = mapper.readValue(new File("tmp-import/" + exerciseToken, "artemis.yaml"), ArtemisExerciseInfo.class);
        ExerciseInfo exerciseInfo = mapper.readValue(new File("tmp-import/" + exerciseToken, "metadata.yaml"), ExerciseInfo.class);
        artemisExerciseInfo.appendExerciseInfo(exerciseInfo);
        return artemisExerciseInfo;
    }

    /**
     * Used to temporary store content of Zip-File into a tmp-subdirectory
     * @param zipInputStream containing the Zip-File to unzip
     * @param subDir name of the tmp-subdirectory
     */
    private void storeZipInTmp(ZipInputStream zipInputStream, String subDir) throws IOException {
        if (!new File("tmp-import").exists()) {
            if (!new File("tmp-import").mkdir()) {
                throw new IOException("Could not create missing '/tmp-import' directory");
            }
        }
        File tmpDir = new File("tmp-import", subDir);
        if (!tmpDir.mkdir()) {
            throw new IOException("Temporary exercise directory could not be created");
        }

        ZipEntry zipEntry;
        byte[] buffer = new byte[1024];

        // https://www.baeldung.com/java-compress-and-uncompress
        while ((zipEntry = zipInputStream.getNextEntry()) != null) {
            if (zipEntry.getName().contains("/.git/")) {    // important as Artemis exercises come in submodules
                continue;
            }
            File entryFile = new File(tmpDir, zipEntry.getName());

            if (zipEntry.isDirectory()) {
                if (!entryFile.mkdir()) {
                    throw new IOException("Failed to create directory " + entryFile);
                }
            } else {
                // fix for Windows-created archives
                File parent = entryFile.getParentFile();
                if (!parent.isDirectory() && !parent.mkdirs()) {
                    throw new IOException("Failed to create directory " + parent);
                }

                FileOutputStream fos = new FileOutputStream(entryFile);
                int len;
                while ((len = zipInputStream.read(buffer)) > 0) {
                    fos.write(buffer, 0, len);
                }
                fos.close();
            }
        }
        fileService.scheduleForDeletion(tmpDir.toPath(), 60);
    }

    /**
     * Utility method used to apply metadata changes to the exercise's 'metadata.yaml' file
     * @param exerciseInfo containing the metadata values
     * @param exerciseToken of the exercise to update
     * @throws IOException
     */
    private void applyExerciseInfoChanges(ArtemisExerciseInfo exerciseInfo, String exerciseToken) throws IOException {
        ObjectMapper mapper = new ObjectMapper(new YAMLFactory()
            .enable(YAMLGenerator.Feature.INDENT_ARRAYS)
            .disable(YAMLGenerator.Feature.WRITE_DOC_START_MARKER));
        mapper.writeValue(new File("tmp-import/" + exerciseToken, "metadata.yaml"), exerciseInfo.getExerciseInfoOnly());
    }

    /**
     * Used to validate apiKeys by checking if they match with
     * one of the registeredConnectors' keys stored in application-(dev|prod).yml
     * @param apiKey the apikey to validate
     * @return true if equal to the local one, false otherwise
     */
    public boolean validate(String apiKey) {
        for (ApplicationProperties.RegisteredConnector registeredConnector : applicationProperties.getRegisteredConnectors()) {
            if (registeredConnector.getAccessToken().equals(apiKey)) {
                log.debug("ApiKey validated correctly against [" + registeredConnector.getUrl() + "]");
                return true;
            }
        }
        return false;
    }
}
